require 'time'
require 'uri'
require 'openssl'
require 'base64'
require 'net/http'
require 'net/http/post/multipart'
require 'json'
require 'mimemagic'

module Ncmb
  class Client
    def initialize(application_key, client_key)
      @application_key = application_key
      @client_key = client_key
      @api_version = '2013-09-01'
      @domain = 'mb.api.cloud.nifty.com'
    end

    def get(path, hash_query = {})
      request(:GET, path, hash_query)
    end

    def post(path, hash_query = {})
      request(:POST, path, hash_query)
    end

    def put(path, hash_query = {})
      request(:PUT, path, hash_query)
    end

    def delete(path)
      request(:DELETE, path)
    end

    def batch(array_request)
      request(:POST, '/batch', requests: array_request)
    end

    def file
      FileStore.new(@application_key, @client_key)
    end

    def data(klass = nil)
      if klass.nil?
        DataStore.new(@application_key, @client_key)
      else
        DataStore.new(@application_key, @client_key, klass)
      end
    end

    private
    def request(method, path, hash_query = {})
      http = Net::HTTP.new(@domain, 443)
      http.use_ssl = true

      path = "/#{@api_version}#{path}"
      current_time = Time.now.utc.iso8601
      signature = sign(method, hash_query, path, current_time)

      if block_given?
        yield(http, path, current_time, signature)
      else
        case method
        when :GET
          query = URI.encode_www_form(hash_query.map { |k, v| [k, v.to_json] })
          path = "#{path}?#{query}"
          res = http.get(path, header(signature, current_time))
          JSON.parse(res.body)
        when :POST
          res = http.post(path, hash_query.to_json, header(signature, current_time))
          JSON.parse(res.body)
        when :PUT
          res = http.put(path, hash_query.to_json, header(signature, current_time))
          JSON.parse(res.body)
        when :DELETE
          res = http.delete(path, header(signature, current_time))
          'successfully deleted.' if res.code == "200"
        else
          raise Exception 'undefined method.'
        end
      end
    # rescue => e
    #   puts e
    end

    def sign(method, hash_query, path, current_time)
      hash = {
        'SignatureMethod' => 'HmacSHA256',
        'SignatureVersion' => 2,
        'X-NCMB-Application-Key' => @application_key,
        'X-NCMB-Timestamp' => current_time
      }

      if method == :GET
        hash_query.each do |key, value|
          hash.store(key.to_s, URI.encode_www_form_component(value.to_json))
        end
      end
      hash = Hash[hash.sort]
      chained_param = hash.map { |k, v| "#{k}=#{v}" }.join('&')
      base_param = [method, @domain, path, chained_param].join("\n")
      signature = Base64.encode64(OpenSSL::HMAC::digest('sha256', @client_key, base_param))
    end

    def header(signature, current_time, content_type = 'application/json')
      {
        'X-NCMB-Application-Key' => @application_key,
        'X-NCMB-Signature' => signature,
        'X-NCMB-Timestamp' => current_time,
        'Content-Type' => content_type,
      }
    end
  end

  class DataStore < Client
    alias :super_get :get

    def initialize(application_key, client_key, klass = nil)
      @klass = klass
      super(application_key, client_key)
    end

    # # #
    # DataStore method

    def post(hash_query = {}, klass = @klass)
      validate_klass klass
      super("/classes/#{klass}", hash_query)
    end

    def get(id, klass = @klass)
      validate_klass klass
      super("/classes/#{klass}/#{id}")
    end

    def put(id, hash_query, klass = @klass)
      validate_klass klass
      super("/classes/#{klass}/#{id}", hash_query)
    end

    def delete(id, klass = @klass)
      validate_klass klass
      super("/classes/#{klass}/#{id}")
    end

    def all(klass = @klass)
      validate_klass klass
      super_get("/classes/#{klass}")
    end

    def where(hash_query, klass = @klass)
      validate_klass klass
      super_get("/classes/#{klass}", where: hash_query)
    end

    def search(hash_query = {}, klass = @klass)
      validate_klass klass
      super_get("/classes/#{klass}", hash_query)
    end

    # ##
    # syntax sugar

    # syntax sugar of search
    def find(hash_query = {}, klass = @klass)
      search(hash_query, klass)
    end

    # syntax sugar of post
    def save(hash_query = {}, klass = @klass)
      post(hash_query, klass)
    end

    # syntax sugar of put
    def update(id, hash_query, klass = @klass)
      put(id, hash_query, klass)
    end

    # syntax sugar of delete
    def destroy(id, klass = @klass)
      delete(id, klass)
    end

    # # #
    # batch method

    # array_query = [
    #   {foo: :bar},
    #   {hoge: :fuga}
    # ]
    def batch_post(array_query)
      validate_klass @klass
      array_request = array_query.map do |query|
        {
          method: :POST,
          path: "#{@api_version}/classes/#{@klass}",
          body: query
        }
      end
      batch_response batch(array_request)
    end

    # array_query = [
    #   {id: 'XXXXXXXX', query: {foo: :baz}},
    #   {id: 'YYYYYYYY', query: {hoge: :piyo}}
    # ]
    # array_query = [
    #   ['XXXXXXXX', {foo: :baz}],
    #   ['YYYYYYYY', {hoge: :piyo}}
    # ]
    def batch_put(array_query)
      validate_klass @klass
      array_request = array_query.map do |obj|
        case obj
        when Array
          {
            method: :PUT,
            path: "#{@api_version}/classes/#{@klass}/#{obj[0]}",
            body: obj[1]
          }
        when Hash
          {
            method: :PUT,
            path: "#{@api_version}/classes/#{@klass}/#{obj[:id]}",
            body: obj[:query]
          }
        else
          {}
        end
      end
      batch_response batch(array_request)
    end

    # array_query = [
    #   'XXXXXXXX',
    #   'YYYYYYYY'
    # ]
    def batch_delete(array_query)
      validate_klass @klass
      array_request = array_query.map do |id|
        {
          method: :DELETE,
          path: "#{@api_version}/classes/#{@klass}/#{id}",
          body: {}
        }
      end
      batch_response batch(array_request)
    end

    private
    def validate_klass(klass)
      raise 'NCMB DataStore Class Name is undefined.' if klass.nil?
    end

    def batch_response(res)
      def res.success
        self.map { |r| r['success'] }.compact
      end

      def res.error
        self.map { |r| r['error'] }.compact
      end

      return res
    end
  end

  class FileStore < Client
    alias :super_get :get

    def post(file, acl = {})
      name = File.basename(file)
      path = "/files/#{name}"
      request(:POST, path, acl) { |http, path, current_time, signature|
        req = Net::HTTP::Post::Multipart.new(
          path, {file: UploadIO.new(file, MimeMagic.by_path(name).to_s)},
          header(signature, current_time, 'multipart/form-data')
        )
        res = http.request(req)
        JSON.parse(res.body)
      }
    end

    def get(name)
      path = "/files/#{name}"
      request(:GET, path) { |http, path, current_time, signature|
        res = http.get(path, header(signature, current_time))
        res.body
      }
    end

    def put(name, acl)
      super("/files/#{name}/", acl: acl)
    end

    def delete(name)
      super("/files/#{name}")
    end

    def all
      super_get('/files')
    end

    # noinspection RubyArgCount
    def where(hash_query)
      super_get('/files', where: hash_query)
    end

    # noinspection RubyArgCount
    def search(hash_query = {})
      super_get('/files', hash_query)
    end
  end
end
